/*========================================================================
                              
                                ** UMOP **
		** Universal Multi-Agent OBDD-based Planner **

            Copyright (C) 1999 by Rune M. Jensen and Manuela M. Veloso
            Computer Science Department, Carnegie Mellon University.
            
    Permission is hereby granted to distribute the UMOP software for 
    research purposes, provided this notice is included with any such 
    distribution. 

      *** WE ARE NOT RESPONSIBLE FOR ANY KIND OF DAMAGE TO YOUR  ***
      *** FILES, DATA, HARDWARE, LOSS OF MONEY, SYSTEM CRASHES   *** 
      *** OR ANY OTHER THING YOU MIGHT COME UP WITH.             ***
      *** - USE THIS PROGRAM OF YOUR OWN FREE WILL !!!           ***

========================================================================*/
/****************************************************** 
 * File  : domain.cc   
 * Desc. : NADL domain member functions
 * Author: Rune M. Jensen
 * Date  : 2/22/99
 ******************************************************/

#include <stream.h>
#include "set.hpp"
#include "domain.hpp"




/****************************************************
 *          print functions                         *
 ****************************************************/

void numberexp::print() {
  switch (type) {
  case nt_number:
    cout << number;
    break;
  case nt_plus:
    cout << "(";
    left->print();
    cout << " + ";
    right->print();
    cout << ")";
    break;
  case nt_minus:
    cout << "(";
    left->print();
    cout << " - ";
    right->print();
    cout << ")";
    break;
  case nt_var:
    cout << var;
    break;
  }
}


void numberprop::print() {
  left->print();
  switch (op) {
  case ro_lt:
    cout << " < ";
    break;
  case ro_gt:
    cout << " > ";
    break;
  case ro_eq:
    cout << " = ";
    break;
  case ro_ne:
    cout << " <> ";
    break;
  }
  right->print();
}  



void atom::print() {
  switch (type) {
  case at_numprop:
    cout << "[";
    prop->print();
    cout << "]";
    break;
  case at_boolvar:
    cout << var;
    break;
  }
}
    


void formula::print() {
  switch (type) {
  case ft_atom:
    atomic->print();
    break;
  case ft_neg:
    cout << "~";
    cout << "{";
    f1->print();
    cout << "}";
    break;
  case ft_and:
    cout << "{";
    f1->print();
    cout << " /\\ ";
    f2->print();
    cout << "}";
    break;
  case ft_or:
    cout << "{";
    f1->print();
    cout << " \\/ ";
    f2->print();
    cout << "}";
    break;
  case ft_impl:
    cout << "{";
    f1->print();
    cout << " => ";
    f2->print();
    cout << "}";
    break;
  case ft_biimpl:
    cout << "{";
    f1->print();
    cout << " <=> ";
    f2->print();
    cout << "}";
    break;
  case ft_ite:
    cout << "{";
    f1->print();
    cout << " -> ";
    f2->print();
    cout << " , ";
    f3->print();
    cout << "}";
    break;
  case ft_true:
    cout << "true";
    break;
  case ft_false:
    cout << "false";
    break;
  case ft_paren:
    cout << "(";
    f1->print();
    cout << ")";
    break;

  }
}    


void action::print() {

  cout << "      " << name << "\n";
  cout << "        con: ";
  strprint(con);
  cout << "\n        pre: ";
  pre->print();
  cout << "\n        eff: ";
  eff->print();
}


void agent::print() {
  cout << "    agt: " << name << "\n\n";
    
  list<action>::iterator i;
  for (i=actions.begin(); i != actions.end(); ++i)
    {
      i->print();
      cout << "\n\n";
    }
}


void vartype::print() {
  if (type == vt_bool)
    cout << "    bool  ";
  else
    cout << "    nat(" << range << ")  ";
}




void vardecl::print() {
  type->print();
  strprint(ids);
  cout << endl;
}



void problem::print() {
  cout << "\nvariables\n\n";
  list<vardecl>::iterator i;
  for (i=vars.begin(); i != vars.end(); ++i) 
    i->print();
  cout << "\n\nsystem\n\n";
  list<agent>::iterator j;
  for (j=sys.begin(); j != sys.end(); ++j) 
    j->print();
  cout << "\n\nenvironment\n\n";
  for (j=env.begin(); j != env.end(); ++j) 
    j->print();
  cout << "\n\ninit\n";
  init->print();
  cout << "\n\ngoal\n";
  goal->print();
  cout << "\n";
}







/****************************************************
 *        functions for converting +- expressions   *
 *        to + expressions                          *     
 ****************************************************/


// IN
//  this : numberprop
// OUT
//  numberprop with all '-' expressions eliminated
void numberprop::minus2plus() {
  numberexp *l,*r,*a,*b,*c,*d;

  left->minus2plus();
  right->minus2plus();
    
    if (left->type == nt_number || left->type == nt_var || left->type == nt_plus)
      {
	if (right->type == nt_number || right->type == nt_var || right->type == nt_plus)
	  {
	    // left: +exp,   right: +exp 
	    // already ok: nothing to do
	  }
	else
	  {
	    // org left: +exp,   right: -exp
	    a = left;
	    b = right->left;
	    r = right;
	    r->type = nt_plus;
	    r->left = a;
	    l = b;
	    left = r;
	    right = l;
	  }
      }
    else
      {
	if (right->type == nt_number || right->type == nt_var || right->type == nt_plus)
	  {
	    // left: -exp,   right: +exp 
	    a = left->left;
	    c = right;
	    l = left;
 	    l->type = nt_plus;
	    l->left = c;
	    r = a;
	    left = r;
	    right = l;
	  }
	else
	  {
	    // left: -exp,   right: -exp 
	    b = left->right;
	    d = right->right;
	    l = left;
	    l->type = nt_plus;
	    l->right = d;
	    r = right;
	    r->type = nt_plus;
	    r->right = b;
	  }
      }
}




// IN
//  this : numberexp
// OUT
//  numberexp with only one top '-' expression
void numberexp::minus2plus() {
  numberexp *l,*r,*a,*b,*c,*d;

  switch (type) {
  case nt_number:
  case nt_var:
    // already ok: do nothing
    break;
  case nt_plus:
    left->minus2plus();
    right->minus2plus();
    if (left->type == nt_number || left->type == nt_var || left->type == nt_plus)
      {
	if (right->type == nt_number || right->type == nt_var || right->type == nt_plus)
	  {
	    // org: +exp, left: +exp,   right: +exp 
	    // already ok: do nothing
	  }
	else
	  {
	    // org +exp, left: +exp,   right: -exp
	    a = left;
	    c = right->right;
	    l = c;
	    r = right;
	    r->right = a;
	    r->type = nt_plus;
            type = nt_minus;
	    left = r;
	    right = l;
	  }
      }
    else
      {
	if (right->type == nt_number || right->type == nt_var || right->type == nt_plus)
	  {
	    // org: +exp, left: -exp,   right: +exp 
	    b = left->right;
	    c = right;
	    l = left;
	    l->right = c;
	    l->type = nt_plus;
	    r = b;
	    type = nt_minus;
	    left = l;
	    right = r;
	  }
	else
	  {
	    // org +exp, left: -exp,   right: -exp
	    b = left->right;
	    c = right->left;
	    l = left;
	    l->type = nt_plus;
	    l->right = c;
	    r = right;
	    r->type = nt_plus;
	    r->left = b;
	    type = nt_minus;
	  }
      }
    break;
  case nt_minus:
    left->minus2plus();
    right->minus2plus();
    if (left->type == nt_number || left->type == nt_var || left->type == nt_plus)
      {
	if (right->type == nt_number || right->type == nt_var || right->type == nt_plus)
	  {
	    // org: -exp, left: +exp,   right: +exp
	    // already ok : do nothing
	  }
	else
	  {
	    // org -exp, left: +exp,   right: -exp
	    a = left;
	    b = right->left;
	    l = b;
	    r = right;
	    r->type = nt_plus;
	    l->left = a;
	    left = r;
	    right = l;
	  }
      }
    else
      {
	if (right->type == nt_number || right->type == nt_var || right->type == nt_plus)
	  {
	    // org: -exp, left: -exp,   right: +exp 
	    a = left->left;
	    c = right;
	    l = left;
	    l->type = nt_plus;
	    l->left = c;
	    r = a;
	    left = r;
	    right = l;
	  }
	else
	  {
	    // org -exp, left: -exp,   right: -exp
	    b = left->right;
	    d = right->right;
	    l = left;
	    l->type = nt_plus;
	    l->right = d;
	    r = right;
	    r->type = nt_plus;
	    r->right = b;
	  }
      }
    break;
  }
}





 
 
