/*========================================================================
                              
                                ** UMOP **
		** Universal Multi-Agent OBDD-based Planner **

            Copyright (C) 1999 by Rune M. Jensen and Manuela M. Veloso
            Computer Science Department, Carnegie Mellon University.
            
    Permission is hereby granted to distribute the UMOP software for 
    research purposes, provided this notice is included with any such 
    distribution. 

      *** WE ARE NOT RESPONSIBLE FOR ANY KIND OF DAMAGE TO YOUR  ***
      *** FILES, DATA, HARDWARE, LOSS OF MONEY, SYSTEM CRASHES   *** 
      *** OR ANY OTHER THING YOU MIGHT COME UP WITH.             ***
      *** - USE THIS PROGRAM OF YOUR OWN FREE WILL !!!           ***

========================================================================*/
/****************************************************** 
 * File  : main.cc   
 * Desc. : main file for the UMOP planner
 * Author: Rune M. Jensen
 * Date  : 4/4/99, 10/27/00
 ******************************************************/
#include <stream.h>
#include <string.h>
#include <stdio.h> 
#include <stdlib.h>
#include <bdd.h>
#include "common.hpp"
#include "domain.hpp"
#include "fsm.hpp"
#include "timer.hpp"
#include "bddprint.hpp"
#include "plan.hpp"
#include "main.hpp"


void  printusage();

int verbosity;

int main(int argc,char **argv)  {
  extern FILE *yyin;
  domdef *def;
  timer watch;
  
  trel *T; 
  bdd sa,init,goal;
  int i,ind;
  char buf[16];
  int res;

  // default values of input parameters
  char domainname[MAXNAMELENGTH] = "domain.nadl";
  char planname[MAXNAMELENGTH] =   "plan.bdd" ;

  int dynreorder = 0;
  int bddnodenum = 100000;
  int cachesize  = 1000;
  int partlimit  = 10000;
  int algorithm  = 0;
 
  verbosity      = 0;

  // initialize and start timer
  resettimers();
  watch.start();

		    
  if (argc < 2) {
    printusage();
    exit (1);
  }

  cout << "Parsing...\n";
  cout.flush();

  /*** Scan command line ***/
  for (ind = 1; ind < argc; ind++) {
    
    /*** Parse switches ***/
    if (argv[ind][0] == '-') {
      switch (argv[ind][1]) {  
      case 'd': strcpy(domainname,argv[++ind]);
	break;
      case 'o': strcpy(planname,argv[++ind]);
	break;
      case 'h': printusage();
	exit(0);
	break;
      case 'a': algorithm  = atoi(argv[++ind]);
	break;
      case 'r': dynreorder = atoi(argv[++ind]);
	break;
      case 't': partlimit  = atoi(argv[++ind]);
	break;
      case 'n': bddnodenum = atoi(argv[++ind]);
	break;
      case 'v': verbosity  = atoi(argv[++ind]);
	break;
      case 'c': cachesize  = atoi(argv[++ind]);
	break;
      default : printf("main.cc main : Unknown switch '%c'\nexiting\n", argv[ind][1]);
	exit(1);
	break;
      }
    }
  }


  // parse umop domain
  yyin = fopen(domainname,"r");
  if (yyin == NULL) 
    {
      cout << "main.cc main : Cannot open \"" << domainname << "\"\n";
      exit(1);
    }


  // exit on syntax error
  if (yyparse()) exit(1); 

  if (verbosity > 1)
    prob->print();
  
  // make domain info. struct (1st add of info to def)
  def = mkdomdef(prob);

  if (verbosity > 1)
    prob->print();
    
  // initialize bdd package
  bdd_init(bddnodenum,cachesize);
  bdd_setvarnum(def->bddvarnum);
        
  // add bdd structures defining variable locations
  def->rels = mkvarrels(def); //(2nd add of info to def)
  
  // enable automatic bdd variable reordering
  switch (dynreorder) {
  case 0:
    mkbddblocks(def);
    bdd_autoreorder(BDD_REORDER_WIN2ITE);
    break;
  case 1:
    mkbddblocks(def);
    bdd_autoreorder(BDD_REORDER_WIN2);
    break;
  case 2:
    // reordering off, nothing to do
    break;
  default : printf("main.cc main : Unknown reorder strategy\nexiting\n");
    exit(1);
    break;
  }
  
  // assign planning algorithm arguments
  sa = bddfalse; 
  init = formula2bdd(prob->init,def);
  goal = formula2bdd(prob->goal,def);

  cout << "Building transition relation...\n";  
  cout.flush();

  // make  partitions
  T = mkpartitions(def);
  ranjanoptimize(T,partlimit); 

/*  
  // debug
  bdd t;
  t = bddtrue;
  for (int p = 0; p < T->partnum; p++)
    t &= T->p[p].exp;
  
  bddprintabs pr("t.bdd",def);
  pr.bddprint(t);
  exit(1);
*/ 
  
  if (verbosity > 1)
    T->print();
  

  cout << "Planning...\n";
  cout.flush();
  
  // ***** select planning algorithm *******
  switch (algorithm) {    
  case 0: // Strong      
    res = SP(T,init,goal,&sa,def);
    break;
  case 1: // Strong Cyclic
    res = SCP(T,init,goal,&sa,def);
    break;
  case 2: // Optimistic
    res = OP(T,init,goal,&sa,def);
    break;
  default: 
    printf("main.cc main : Unknown planning algorithm\nexiting\n");
    exit(1);
    break;
  }
  
  if (res)
    cout << "Universal plan covers Init\n";    
  else
    cout << "Universal plan does not cover Init\n";
  
  // print elapsed time
  watch.stop();
  
  // write bdd output
  cout << "Final node count of universal plan: " << bdd_nodecount(sa) << "\n";
  cout << "Write universal plan to output file \"" << planname << "\" (y/n): ";
  cin >> buf;
  if (buf[0] == 'y')
    {
      bddprintabs p(planname,def);
      p.bddprint(sa);
    }
  
  
  // close bdd package
  bdd_done();
  
  return 0; 
}





void  printusage() {
  printf("USAGE:\n");
  printf("       -d <domain file>.nadl    (default \"domain.nadl\")\n");
  printf("       -o <planfile>.out        (default \"plan.out\")\n");
  printf("       -t partition threshold   (default 10000)\n");
  printf("       -n bdd node count        (default 100000)\n");
  printf("       -c bdd cache size        (default 10000)\n");
  printf("       -v verbosity level       (default 0)\n"); 
  printf("       -h write usage                           \n");
  printf("       -a algorithm             (default 0)\n");
  printf("           0: Strong\n");
  printf("           1: StrongCyclic\n");
  printf("           2: Optimistic\n");
  printf("       -r dynamic reorder alg.  (default 0)\n");
  printf("           0: WIN2ITE\n");
  printf("           1: WIN2\n");
  printf("           2: reordering off\n");
  printf("\n");
}

