/*========================================================================
                              
                                ** UMOP **
		** Universal Multi-Agent OBDD-based Planner **

            Copyright (C) 1999 by Rune M. Jensen and Manuela M. Veloso
            Computer Science Department, Carnegie Mellon University.
            
    Permission is hereby granted to distribute the UMOP software for 
    research purposes, provided this notice is included with any such 
    distribution. 

      *** WE ARE NOT RESPONSIBLE FOR ANY KIND OF DAMAGE TO YOUR  ***
      *** FILES, DATA, HARDWARE, LOSS OF MONEY, SYSTEM CRASHES   *** 
      *** OR ANY OTHER THING YOU MIGHT COME UP WITH.             ***
      *** - USE THIS PROGRAM OF YOUR OWN FREE WILL !!!           ***

========================================================================*/
/****************************************************** 
 * File  : plan.cc  
 * Desc. : Strong StrongCyclic and Optimistic  planning 
 *         algorithms. 
 * Author: Rune M. Jensen CS, CMU
 * Date  : 4/4/99, 5/25/00, 10/26/00
 ******************************************************/

#include <bdd.h>
#include "common.hpp"
#include "domain.hpp"
#include "fsm.hpp"
#include "plan.hpp"
#include "bddprint.hpp"



/****************************************************** 
 *      Aux. functions   
 *     
 ******************************************************/


bdd prunestates(bdd preimage, bdd acc)  {
  return preimage & !acc;
}


bdd projact(bdd sa,domdef *def)  {
  return bdd_exist(sa, def->rels->sysactions);
}


// IN
//  tr   : Partitioned transition relation
//  acc  : currently visited states (encoded in current state variables)
//  rels : bdd variable info structure 
// OUT
//  strongpreimage of acc encoded in act * current state vars
//  the calculation combines a partitioned representation 
bdd strongpreimage(trel *tr, bdd acc,domdef *def)  {
  bdd p1,p2,c;
  int i;

  c = bdd_replace(acc,def->rels->old2new);

  p1 = c;
  // P1 preimage composition 
  for (i=0; i < tr->partnum; i++)
    {
      p1 = bdd_appex(tr->p[i].exp,p1,bddop_and,tr->p[i].varset);
      if (verbosity > 0)
	cout << "Finished Strong partition A" << i << " node count: " 
	     << bdd_nodecount(p1) << "\n";
    }

  p2 = !c;
  // P2 preimage composition 
  for (i=0; i< tr->partnum; i++)
    {
      p2 = bdd_appex(tr->p[i].exp,p2,bddop_and,tr->p[i].varset);
      if (verbosity > 0)
	cout << "Finished Strong partition B" << i << " node count: " 
	     << bdd_nodecount(p2) << "\n";
    }

  return p1 & !p2;
}


// IN
//  tr   : Partitioned transition relation
//  acc  : currently visited states (encoded in current state variables)
//  rels : bdd variable info structure 
// OUT
//  weak preimage of acc encoded in act * current state vars
//  the calculation combines a partitioned representation 
bdd weakpreimage(trel *tr, bdd acc,domdef *def)  {
  bdd res;
  int i;

  res = bdd_replace(acc,def->rels->old2new); 
  // preimage composition 
  for (i=0; i<tr->partnum; i++)
    {
      res = bdd_appex(tr->p[i].exp,res,bddop_and,tr->p[i].varset);
      if (verbosity > 0)
	cout << "Finished Weak partition " << i << " node count: " 
	     << bdd_nodecount(res) << "\n";
    }
  return res;
}





/****************************************************** 
 *          SP  
 *          Strong Planning
 ******************************************************/
// IN
//  tr     : Partitioned transition relation
//  init   : Initial state (encoded in current state vars)
//  goal   : Goal state (encoded in current state vars)
//  sa     : plan output handle
//  def    : Domain definition (containing a description of the 
//           location of variables in bdds.
// OUT
//  1:  some plan found
//  0:  fix point found, but Init is not a subset 
int SP(trel *tr, bdd init, bdd goal, bdd *sa,domdef *def)
{
  bdd acc = goal;                // acc encoded using new state vars (visited states)
  bdd preimage, prunedpreimage; 
  int i = 0;
  
  
  //main loop
  while ((acc & init) != init)
    {
      preimage = strongpreimage(tr,acc,def); // act*current bdd
      prunedpreimage = prunestates(preimage,acc); // act*current bdd

      
      if (prunedpreimage != bddfalse)
	{
	  *sa |= prunedpreimage;
	  acc |= projact(prunedpreimage,def);
	  if (verbosity > 0)
	    cout << "Added preimage " << i << "\n";
	  i++;
	}
      else
	// fix point found, but init not included
	return 0;
    }
  // init is included
  return 1;
}



/****************************************************** 
 *          SCP  
 *          Strong Cyclic Planning
 ******************************************************/


bdd OSA(trel *tr,bdd CSA,bdd s,domdef *def) {
  bdd res,c;
  int i,partnum;
  varrels *rels;

  partnum = tr->partnum;
  rels = def->rels;

  c = bdd_replace(s,rels->old2new);

  res = !c;

  // OSA composition 
  for (i=0; i<partnum; i++)
    {
      res = bdd_appex(tr->p[i].exp,res,bddop_and,tr->p[i].varset);
      if (verbosity > 0)
	cout << "OSA Finished partition " << i << " node count: " 
	     << bdd_nodecount(res) << "\n";
    }

  return res & CSA;
}
  



// IN
//  tr     : Partitioned Transition relation 
//  acc    : Visited states from goal
//  sa     : current s->a table
//  def    : Domain definition (containing a description of the 
//           location of variables in bdds.)
// OUT
//  update of SA and acc
void faircycles(trel *tr, bdd *acc, bdd *sa,bdd *oldacc, domdef *def) {
  varrels *rels;
  bdd IWpi, Outgoing;

  bdd WpiAcc = bddfalse;
  bdd OldWpiAcc = bddtrue;
  bdd CSA = bddfalse;

  rels = def->rels;

  while (CSA == bddfalse && OldWpiAcc != WpiAcc)
    {
      IWpi = weakpreimage(tr,projact(WpiAcc,def) | *acc,def);
      CSA = prunestates(IWpi,*acc);
      Outgoing = CSA;
      while (CSA != bddfalse && Outgoing != bddfalse)
	{
	  Outgoing = OSA(tr,CSA,projact(CSA,def) | *acc,def); 
	  CSA &= !Outgoing;
	}
      if (CSA != bddfalse)
	{
	  *sa |= CSA;
	  *oldacc = *acc;
	  *acc |= projact(CSA,def);
	  return;
	}
      else
	{
	  OldWpiAcc = WpiAcc;
	  WpiAcc |= IWpi;
	} 
    }
  *oldacc = *acc;
}






// IN
//  tr     : Partitioned transition relation
//  init   : Initial state (encoded in current state vars)
//  goal   : Goal state    (encoded in current state vars)
//  sa     : plan output handle
//  def    : Domain definition 
// OUT
//  1:  some plan found
//  0:  no plan found 
int SCP(trel *tr, bdd init, bdd goal, bdd *sa,domdef *def) {
  bdd acc = goal;               // acc encoded using new state vars (visited states)
  bdd oldacc = bddfalse;        // old acc
  bdd preimage, prunedpreimage; 
  int i,nc;

  // make universal plan by backward chaining from goal
  i = 0; // preimage count
  
  //main loop
  while (acc != oldacc)
    {
      if ( (acc & init) == init ) 
	return 1;
      preimage = strongpreimage(tr,acc,def); // act*current bdd
      prunedpreimage = prunestates(preimage,acc); // act*current bdd

      if (prunedpreimage != bddfalse)
	{
	  *sa |= prunedpreimage;
	  oldacc = acc;
	  acc |= projact(prunedpreimage,def);
	  i++;
	}
      else
	{
	  faircycles(tr,&acc,sa,&oldacc,def);
	  i++;
	}
    }
  return 0;
}




/****************************************************** 
 *          OP  
 *          Optimistic planning algorithm
 ******************************************************/


// IN
//  tr     : Partitioned transition relation
//  init   : Initial state (encoded in current state vars)
//  goal   : Goal state    (encoded in current state vars)
//  sa     : plan in act * current-state 
//  def    : Domain definition 
// OUT
//  plan ignoring nondeterminism (uses the real preimage in back chaining)
//  1:  init subset of SA (plan found)
//  0:  init not subset of SA (no plan found)
int OP(trel *tr, bdd init, bdd goal, bdd *sa,domdef *def) {
  bdd acc = goal;               // acc encoded using new state vars (visited states)
  bdd preimage, prunedpreimage; 
 
  // make universal plan by backward chaining from goal
  while ( (acc & init) != init ) 
    {
      preimage = weakpreimage(tr,acc,def); // act*current bdd
      prunedpreimage = prunestates(preimage,acc); // act*current bdd

      if (prunedpreimage != bddfalse)
	{
	  *sa |= prunedpreimage;
	  acc |= projact(prunedpreimage,def);
	}
      else
	// fix point found, but init not subset of acc
	return 0;
    }
  // init subset of acc
  return 1;
}












